@@ -257,6 +257,7 @@ bad_frame:
 static int rtp2trau_fr(struct osmo_trau_frame *tf, const uint8_t *data, size_t data_len)
 {
 	int i, j, k, l, o;
+	enum osmo_gsm631_sid_class sidc;
 
 	/* data_len == 0 for BFI frame */
 	if (data_len < GSM_FR_BYTES && data_len != 0)
@@ -269,44 +270,63 @@ static int rtp2trau_fr(struct osmo_trau_frame *tf, const uint8_t *data, size_t d
 
 	/* FR Data Bits according to TS 48.060 Section 5.5.1.1.2 */
 
-	/* FIXME: Generate SID frames? */
-
 	/* set c-bits and t-bits */
 	if (tf->dir == OSMO_TRAU_DIR_UL) {
-		/* C1 .. C5 */
+		/* C1 .. C5: FR UL */
 		tf->c_bits[0] = 0;
 		tf->c_bits[1] = 0;
 		tf->c_bits[2] = 0;
 		tf->c_bits[3] = 1;
 		tf->c_bits[4] = 0;
-	} else {
-		/* C1 .. C5 */
-		tf->c_bits[0] = 1;
-		tf->c_bits[1] = 1;
-		tf->c_bits[2] = 1;
-		tf->c_bits[3] = 0;
-		tf->c_bits[4] = 0;
+	} else {			/* DL */
+		if (data_len == 0) {
+			/* C1 .. C5: idle speech */
+			tf->c_bits[0] = 0;
+			tf->c_bits[1] = 1;
+			tf->c_bits[2] = 1;
+			tf->c_bits[3] = 1;
+			tf->c_bits[4] = 0;
+		} else {
+			/* C1 .. C5: FR DL */
+			tf->c_bits[0] = 1;
+			tf->c_bits[1] = 1;
+			tf->c_bits[2] = 1;
+			tf->c_bits[3] = 0;
+			tf->c_bits[4] = 0;
+		}
 	}
 	memset(&tf->c_bits[5], 0, 6);	/* C6 .. C11: Time Alignment */
 	if (tf->dir == OSMO_TRAU_DIR_UL) {
-		if (data_len == 0)
+		if (data_len == 0) {
 			tf->c_bits[11] = 1;	/* C12: BFI */
-		else
+			tf->c_bits[12] = 0;	/* C13: SID=0 */
+			tf->c_bits[13] = 0;	/* C14: SID=0 */
+		} else {
 			tf->c_bits[11] = 0;	/* C12: BFI */
-		tf->c_bits[12] = 0; /* C13: SID=0 */
-		tf->c_bits[13] = 0; /* C14: SID=0 */
+			/* SID classification per GSM 06.31 section 6.1.1 */
+			sidc = osmo_fr_sid_classify(data);
+			tf->c_bits[12] = (sidc >> 1) & 1; /* C13: msb */
+			tf->c_bits[13] = (sidc >> 0) & 1; /* C14: lsb */
+		}
 		tf->c_bits[14] = 0; /* C15: TAF (SACCH or not) */
 		tf->c_bits[15] = 1; /* C16: spare */
 		tf->c_bits[16] = 0; /* C17: DTXd not applied */
 	} else {
-		memset(&tf->c_bits[11], 1, 10); /* C12 .. C15: spare */
-		tf->c_bits[15] = 1; /* C16: SP=1 */
+		memset(&tf->c_bits[11], 1, 4);	/* C12 .. C15: spare */
+		if (data_len && osmo_fr_check_sid(data, data_len))
+			tf->c_bits[15] = 0;	/* C16: SP=0 */
+		else
+			tf->c_bits[15] = 1;	/* C16: SP=1 */
+		tf->c_bits[16] = 1; /* C17: spare */
 	}
-	memset(&tf->c_bits[17], 1, 4); /* C18 .. C12: spare */
+	memset(&tf->c_bits[17], 1, 4); /* C18 .. C21: spare */
 	memset(&tf->t_bits[0], 1, 4);
 
-	if (!data_len)
+	if (!data_len) {
+		/* idle speech frame if DL, BFI speech frame if UL */
+		memset(&tf->d_bits[0], 1, 260);
 		return 0;
+	}
 
 	/* reassemble d-bits */
 	i = 0; /* counts bits */
@@ -415,6 +435,7 @@ static int rtp2trau_efr(struct osmo_trau_frame *tf, const uint8_t *data, size_t
 {
 	int i, j;
 	ubit_t check_bits[26];
+	enum osmo_gsm631_sid_class sidc;
 
 	/* data_len == 0 for BFI frame */
 	if (data_len < GSM_EFR_BYTES && data_len != 0)
@@ -428,35 +449,55 @@ static int rtp2trau_efr(struct osmo_trau_frame *tf, const uint8_t *data, size_t
 	/* FR Data Bits according to TS 48.060 Section 5.5.1.1.2 */
 
 	/* set c-bits and t-bits */
-	tf->c_bits[0] = 1;
-	tf->c_bits[1] = 1;
-	tf->c_bits[2] = 0;
-	tf->c_bits[3] = 1;
-	tf->c_bits[4] = 0;
+	if (data_len == 0 && tf->dir == OSMO_TRAU_DIR_DL) {
+		/* C1 .. C5: idle speech */
+		tf->c_bits[0] = 0;
+		tf->c_bits[1] = 1;
+		tf->c_bits[2] = 1;
+		tf->c_bits[3] = 1;
+		tf->c_bits[4] = 0;
+	} else {
+		/* C1 .. C5: EFR */
+		tf->c_bits[0] = 1;
+		tf->c_bits[1] = 1;
+		tf->c_bits[2] = 0;
+		tf->c_bits[3] = 1;
+		tf->c_bits[4] = 0;
+	}
 
 	memset(&tf->c_bits[5], 0, 6); /* C6 .. C11: Time Alignment */
 	if (tf->dir == OSMO_TRAU_DIR_UL) {
-		if (data_len == 0)
-			tf->c_bits[11] = 1; /* C12: BFI=1 */
-		else
-			tf->c_bits[11] = 0; /* C12: BFI=1 */
-		/* FIXME: Generate SID frames? */
-		tf->c_bits[12] = 0; /* C13: SID=0 */
-		tf->c_bits[13] = 0; /* C14: SID=0 */
+		if (data_len == 0) {
+			tf->c_bits[11] = 1;	/* C12: BFI */
+			tf->c_bits[12] = 0;	/* C13: SID=0 */
+			tf->c_bits[13] = 0;	/* C14: SID=0 */
+		} else {
+			tf->c_bits[11] = 0;	/* C12: BFI */
+			/* SID classification per GSM 06.81 section 6.1.1 */
+			sidc = osmo_efr_sid_classify(data);
+			tf->c_bits[12] = (sidc >> 1) & 1; /* C13: msb */
+			tf->c_bits[13] = (sidc >> 0) & 1; /* C14: lsb */
+		}
 		tf->c_bits[14] = 0; /* C15: TAF (SACCH) */
 		tf->c_bits[15] = 1; /* C16: spare */
 		tf->c_bits[16] = 0; /* C17: DTXd applied */
 	} else {
 		tf->c_bits[11] = 1; /* C12: UFE (good uplink) */
 		memset(&tf->c_bits[12], 1, 3);	/* C13 .. C15: spare */
-		tf->c_bits[15] = 1; /* C16: SP=1 */
+		if (data_len && osmo_efr_check_sid(data, data_len))
+			tf->c_bits[15] = 0;	/* C16: SP=0 */
+		else
+			tf->c_bits[15] = 1;	/* C16: SP=1 */
 		tf->c_bits[16] = 1; /* C17: spare */
 	}
 	memset(&tf->c_bits[17], 1, 4);	/* C18 .. C21: spare */
 	memset(&tf->t_bits[0], 1, 4);
 
-	if (data_len == 0)
+	if (data_len == 0) {
+		/* idle speech frame if DL, BFI speech frame if UL */
+		memset(&tf->d_bits[0], 1, 260);
 		return 0;
+	}
 
 	/* reassemble d-bits */
 	tf->d_bits[0] = 1;
